from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MergePolicy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NEVER: _ClassVar[MergePolicy]
    EQUAL: _ClassVar[MergePolicy]
    RETAIN: _ClassVar[MergePolicy]
    REPLACE: _ClassVar[MergePolicy]

class SubscriptionUpdateOp(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ADD_UUIDS: _ClassVar[SubscriptionUpdateOp]
    REMOVE_UUIDS: _ClassVar[SubscriptionUpdateOp]
NEVER: MergePolicy
EQUAL: MergePolicy
RETAIN: MergePolicy
REPLACE: MergePolicy
ADD_UUIDS: SubscriptionUpdateOp
REMOVE_UUIDS: SubscriptionUpdateOp

class RawValuesParams(_message.Message):
    __slots__ = ("uuid", "start", "end", "versionMajor")
    UUID_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    start: int
    end: int
    versionMajor: int
    def __init__(self, uuid: _Optional[bytes] = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., versionMajor: _Optional[int] = ...) -> None: ...

class RawValuesResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "values")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    values: _containers.RepeatedCompositeFieldContainer[RawPoint]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., values: _Optional[_Iterable[_Union[RawPoint, _Mapping]]] = ...) -> None: ...

class ArrowRawValuesParams(_message.Message):
    __slots__ = ("uuid", "start", "end", "versionMajor", "templateBytes")
    UUID_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    TEMPLATEBYTES_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    start: int
    end: int
    versionMajor: int
    templateBytes: bytes
    def __init__(self, uuid: _Optional[bytes] = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., versionMajor: _Optional[int] = ..., templateBytes: _Optional[bytes] = ...) -> None: ...

class ArrowRawValuesResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "arrowBytes")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    arrowBytes: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...

class ArrowMultiValuesParams(_message.Message):
    __slots__ = ("uuid", "versionMajor", "start", "end", "snapPeriodNs", "templateBytes")
    UUID_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    SNAPPERIODNS_FIELD_NUMBER: _ClassVar[int]
    TEMPLATEBYTES_FIELD_NUMBER: _ClassVar[int]
    uuid: _containers.RepeatedScalarFieldContainer[bytes]
    versionMajor: _containers.RepeatedScalarFieldContainer[int]
    start: int
    end: int
    snapPeriodNs: int
    templateBytes: bytes
    def __init__(self, uuid: _Optional[_Iterable[bytes]] = ..., versionMajor: _Optional[_Iterable[int]] = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., snapPeriodNs: _Optional[int] = ..., templateBytes: _Optional[bytes] = ...) -> None: ...

class ArrowMultiValuesResponse(_message.Message):
    __slots__ = ("stat", "arrowBytes")
    STAT_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    arrowBytes: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...

class RawPointVec(_message.Message):
    __slots__ = ("time", "value")
    TIME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    time: int
    value: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, time: _Optional[int] = ..., value: _Optional[_Iterable[float]] = ...) -> None: ...

class AlignedWindowsParams(_message.Message):
    __slots__ = ("uuid", "start", "end", "versionMajor", "pointWidth")
    UUID_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    POINTWIDTH_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    start: int
    end: int
    versionMajor: int
    pointWidth: int
    def __init__(self, uuid: _Optional[bytes] = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., versionMajor: _Optional[int] = ..., pointWidth: _Optional[int] = ...) -> None: ...

class AlignedWindowsResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "values")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    values: _containers.RepeatedCompositeFieldContainer[StatPoint]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., values: _Optional[_Iterable[_Union[StatPoint, _Mapping]]] = ...) -> None: ...

class ArrowAlignedWindowsResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "arrowBytes")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    arrowBytes: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...

class WindowsParams(_message.Message):
    __slots__ = ("uuid", "start", "end", "versionMajor", "width", "depth")
    UUID_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    WIDTH_FIELD_NUMBER: _ClassVar[int]
    DEPTH_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    start: int
    end: int
    versionMajor: int
    width: int
    depth: int
    def __init__(self, uuid: _Optional[bytes] = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., versionMajor: _Optional[int] = ..., width: _Optional[int] = ..., depth: _Optional[int] = ...) -> None: ...

class WindowsResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "values")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    values: _containers.RepeatedCompositeFieldContainer[StatPoint]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., values: _Optional[_Iterable[_Union[StatPoint, _Mapping]]] = ...) -> None: ...

class ArrowWindowsResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "arrowBytes")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    arrowBytes: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...

class StreamInfoParams(_message.Message):
    __slots__ = ("uuid", "omitVersion", "omitDescriptor", "role")
    UUID_FIELD_NUMBER: _ClassVar[int]
    OMITVERSION_FIELD_NUMBER: _ClassVar[int]
    OMITDESCRIPTOR_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    omitVersion: bool
    omitDescriptor: bool
    role: Role
    def __init__(self, uuid: _Optional[bytes] = ..., omitVersion: bool = ..., omitDescriptor: bool = ..., role: _Optional[_Union[Role, _Mapping]] = ...) -> None: ...

class StreamInfoResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "descriptor")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTOR_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    descriptor: StreamDescriptor
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., descriptor: _Optional[_Union[StreamDescriptor, _Mapping]] = ...) -> None: ...

class StreamDescriptor(_message.Message):
    __slots__ = ("uuid", "collection", "tags", "annotations", "propertyVersion")
    UUID_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    PROPERTYVERSION_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    collection: str
    tags: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    annotations: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    propertyVersion: int
    def __init__(self, uuid: _Optional[bytes] = ..., collection: _Optional[str] = ..., tags: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., annotations: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., propertyVersion: _Optional[int] = ...) -> None: ...

class SetStreamAnnotationsParams(_message.Message):
    __slots__ = ("uuid", "expectedPropertyVersion", "changes", "removals")
    UUID_FIELD_NUMBER: _ClassVar[int]
    EXPECTEDPROPERTYVERSION_FIELD_NUMBER: _ClassVar[int]
    CHANGES_FIELD_NUMBER: _ClassVar[int]
    REMOVALS_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    expectedPropertyVersion: int
    changes: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    removals: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, uuid: _Optional[bytes] = ..., expectedPropertyVersion: _Optional[int] = ..., changes: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., removals: _Optional[_Iterable[str]] = ...) -> None: ...

class SetStreamAnnotationsResponse(_message.Message):
    __slots__ = ("stat",)
    STAT_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class SetStreamTagsParams(_message.Message):
    __slots__ = ("uuid", "expectedPropertyVersion", "tags", "collection", "remove")
    UUID_FIELD_NUMBER: _ClassVar[int]
    EXPECTEDPROPERTYVERSION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    REMOVE_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    expectedPropertyVersion: int
    tags: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    collection: str
    remove: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, uuid: _Optional[bytes] = ..., expectedPropertyVersion: _Optional[int] = ..., tags: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., collection: _Optional[str] = ..., remove: _Optional[_Iterable[str]] = ...) -> None: ...

class SetStreamTagsResponse(_message.Message):
    __slots__ = ("stat",)
    STAT_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class CreateParams(_message.Message):
    __slots__ = ("uuid", "collection", "tags", "annotations")
    UUID_FIELD_NUMBER: _ClassVar[int]
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    collection: str
    tags: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    annotations: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    def __init__(self, uuid: _Optional[bytes] = ..., collection: _Optional[str] = ..., tags: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., annotations: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ...) -> None: ...

class CreateResponse(_message.Message):
    __slots__ = ("stat",)
    STAT_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class MetadataUsageParams(_message.Message):
    __slots__ = ("prefix", "role")
    PREFIX_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    prefix: str
    role: Role
    def __init__(self, prefix: _Optional[str] = ..., role: _Optional[_Union[Role, _Mapping]] = ...) -> None: ...

class MetadataUsageResponse(_message.Message):
    __slots__ = ("stat", "tags", "annotations")
    STAT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    tags: _containers.RepeatedCompositeFieldContainer[KeyCount]
    annotations: _containers.RepeatedCompositeFieldContainer[KeyCount]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., tags: _Optional[_Iterable[_Union[KeyCount, _Mapping]]] = ..., annotations: _Optional[_Iterable[_Union[KeyCount, _Mapping]]] = ...) -> None: ...

class KeyCount(_message.Message):
    __slots__ = ("key", "count")
    KEY_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    key: str
    count: int
    def __init__(self, key: _Optional[str] = ..., count: _Optional[int] = ...) -> None: ...

class ListCollectionsParams(_message.Message):
    __slots__ = ("prefix", "role")
    PREFIX_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    prefix: str
    role: Role
    def __init__(self, prefix: _Optional[str] = ..., role: _Optional[_Union[Role, _Mapping]] = ...) -> None: ...

class ListCollectionsResponse(_message.Message):
    __slots__ = ("stat", "collections")
    STAT_FIELD_NUMBER: _ClassVar[int]
    COLLECTIONS_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    collections: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., collections: _Optional[_Iterable[str]] = ...) -> None: ...

class LookupStreamsParams(_message.Message):
    __slots__ = ("collection", "isCollectionPrefix", "tags", "annotations", "role")
    COLLECTION_FIELD_NUMBER: _ClassVar[int]
    ISCOLLECTIONPREFIX_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    collection: str
    isCollectionPrefix: bool
    tags: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    annotations: _containers.RepeatedCompositeFieldContainer[KeyOptValue]
    role: Role
    def __init__(self, collection: _Optional[str] = ..., isCollectionPrefix: bool = ..., tags: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., annotations: _Optional[_Iterable[_Union[KeyOptValue, _Mapping]]] = ..., role: _Optional[_Union[Role, _Mapping]] = ...) -> None: ...

class LookupStreamsResponse(_message.Message):
    __slots__ = ("stat", "results")
    STAT_FIELD_NUMBER: _ClassVar[int]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    results: _containers.RepeatedCompositeFieldContainer[StreamDescriptor]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., results: _Optional[_Iterable[_Union[StreamDescriptor, _Mapping]]] = ...) -> None: ...

class NearestParams(_message.Message):
    __slots__ = ("uuid", "time", "versionMajor", "backward")
    UUID_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    BACKWARD_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    time: int
    versionMajor: int
    backward: bool
    def __init__(self, uuid: _Optional[bytes] = ..., time: _Optional[int] = ..., versionMajor: _Optional[int] = ..., backward: bool = ...) -> None: ...

class NearestResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "value")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    value: RawPoint
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., value: _Optional[_Union[RawPoint, _Mapping]] = ...) -> None: ...

class ChangesParams(_message.Message):
    __slots__ = ("uuid", "fromMajor", "toMajor", "resolution")
    UUID_FIELD_NUMBER: _ClassVar[int]
    FROMMAJOR_FIELD_NUMBER: _ClassVar[int]
    TOMAJOR_FIELD_NUMBER: _ClassVar[int]
    RESOLUTION_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    fromMajor: int
    toMajor: int
    resolution: int
    def __init__(self, uuid: _Optional[bytes] = ..., fromMajor: _Optional[int] = ..., toMajor: _Optional[int] = ..., resolution: _Optional[int] = ...) -> None: ...

class ChangesResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor", "ranges")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    RANGES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    ranges: _containers.RepeatedCompositeFieldContainer[ChangedRange]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ..., ranges: _Optional[_Iterable[_Union[ChangedRange, _Mapping]]] = ...) -> None: ...

class RoundSpec(_message.Message):
    __slots__ = ("bits",)
    BITS_FIELD_NUMBER: _ClassVar[int]
    bits: int
    def __init__(self, bits: _Optional[int] = ...) -> None: ...

class InsertParams(_message.Message):
    __slots__ = ("uuid", "sync", "merge_policy", "rounding", "values")
    UUID_FIELD_NUMBER: _ClassVar[int]
    SYNC_FIELD_NUMBER: _ClassVar[int]
    MERGE_POLICY_FIELD_NUMBER: _ClassVar[int]
    ROUNDING_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    sync: bool
    merge_policy: MergePolicy
    rounding: RoundSpec
    values: _containers.RepeatedCompositeFieldContainer[RawPoint]
    def __init__(self, uuid: _Optional[bytes] = ..., sync: bool = ..., merge_policy: _Optional[_Union[MergePolicy, str]] = ..., rounding: _Optional[_Union[RoundSpec, _Mapping]] = ..., values: _Optional[_Iterable[_Union[RawPoint, _Mapping]]] = ...) -> None: ...

class ArrowInsertParams(_message.Message):
    __slots__ = ("uuid", "sync", "merge_policy", "rounding", "arrowBytes")
    UUID_FIELD_NUMBER: _ClassVar[int]
    SYNC_FIELD_NUMBER: _ClassVar[int]
    MERGE_POLICY_FIELD_NUMBER: _ClassVar[int]
    ROUNDING_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    sync: bool
    merge_policy: MergePolicy
    rounding: RoundSpec
    arrowBytes: bytes
    def __init__(self, uuid: _Optional[bytes] = ..., sync: bool = ..., merge_policy: _Optional[_Union[MergePolicy, str]] = ..., rounding: _Optional[_Union[RoundSpec, _Mapping]] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...

class InsertResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ...) -> None: ...

class DeleteParams(_message.Message):
    __slots__ = ("uuid", "start", "end")
    UUID_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    start: int
    end: int
    def __init__(self, uuid: _Optional[bytes] = ..., start: _Optional[int] = ..., end: _Optional[int] = ...) -> None: ...

class DeleteResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ...) -> None: ...

class InfoParams(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class InfoResponse(_message.Message):
    __slots__ = ("stat", "mash", "majorVersion", "minorVersion", "build", "proxy")
    STAT_FIELD_NUMBER: _ClassVar[int]
    MASH_FIELD_NUMBER: _ClassVar[int]
    MAJORVERSION_FIELD_NUMBER: _ClassVar[int]
    MINORVERSION_FIELD_NUMBER: _ClassVar[int]
    BUILD_FIELD_NUMBER: _ClassVar[int]
    PROXY_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    mash: Mash
    majorVersion: int
    minorVersion: int
    build: str
    proxy: ProxyInfo
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., mash: _Optional[_Union[Mash, _Mapping]] = ..., majorVersion: _Optional[int] = ..., minorVersion: _Optional[int] = ..., build: _Optional[str] = ..., proxy: _Optional[_Union[ProxyInfo, _Mapping]] = ...) -> None: ...

class ProxyInfo(_message.Message):
    __slots__ = ("proxyEndpoints",)
    PROXYENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    proxyEndpoints: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, proxyEndpoints: _Optional[_Iterable[str]] = ...) -> None: ...

class FaultInjectParams(_message.Message):
    __slots__ = ("type", "params")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    type: int
    params: bytes
    def __init__(self, type: _Optional[int] = ..., params: _Optional[bytes] = ...) -> None: ...

class FaultInjectResponse(_message.Message):
    __slots__ = ("stat", "rv")
    STAT_FIELD_NUMBER: _ClassVar[int]
    RV_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    rv: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., rv: _Optional[bytes] = ...) -> None: ...

class FlushParams(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    def __init__(self, uuid: _Optional[bytes] = ...) -> None: ...

class FlushResponse(_message.Message):
    __slots__ = ("stat", "versionMajor", "versionMinor")
    STAT_FIELD_NUMBER: _ClassVar[int]
    VERSIONMAJOR_FIELD_NUMBER: _ClassVar[int]
    VERSIONMINOR_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    versionMajor: int
    versionMinor: int
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., versionMajor: _Optional[int] = ..., versionMinor: _Optional[int] = ...) -> None: ...

class ObliterateParams(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    def __init__(self, uuid: _Optional[bytes] = ...) -> None: ...

class ObliterateResponse(_message.Message):
    __slots__ = ("stat",)
    STAT_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class RawPoint(_message.Message):
    __slots__ = ("time", "value")
    TIME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    time: int
    value: float
    def __init__(self, time: _Optional[int] = ..., value: _Optional[float] = ...) -> None: ...

class StatPoint(_message.Message):
    __slots__ = ("time", "min", "mean", "max", "count", "stddev")
    TIME_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MEAN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    STDDEV_FIELD_NUMBER: _ClassVar[int]
    time: int
    min: float
    mean: float
    max: float
    count: int
    stddev: float
    def __init__(self, time: _Optional[int] = ..., min: _Optional[float] = ..., mean: _Optional[float] = ..., max: _Optional[float] = ..., count: _Optional[int] = ..., stddev: _Optional[float] = ...) -> None: ...

class ChangedRange(_message.Message):
    __slots__ = ("start", "end")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    start: int
    end: int
    def __init__(self, start: _Optional[int] = ..., end: _Optional[int] = ...) -> None: ...

class Status(_message.Message):
    __slots__ = ("code", "msg", "mash")
    CODE_FIELD_NUMBER: _ClassVar[int]
    MSG_FIELD_NUMBER: _ClassVar[int]
    MASH_FIELD_NUMBER: _ClassVar[int]
    code: int
    msg: str
    mash: Mash
    def __init__(self, code: _Optional[int] = ..., msg: _Optional[str] = ..., mash: _Optional[_Union[Mash, _Mapping]] = ...) -> None: ...

class Mash(_message.Message):
    __slots__ = ("revision", "leader", "leaderRevision", "totalWeight", "healthy", "unmapped", "members")
    REVISION_FIELD_NUMBER: _ClassVar[int]
    LEADER_FIELD_NUMBER: _ClassVar[int]
    LEADERREVISION_FIELD_NUMBER: _ClassVar[int]
    TOTALWEIGHT_FIELD_NUMBER: _ClassVar[int]
    HEALTHY_FIELD_NUMBER: _ClassVar[int]
    UNMAPPED_FIELD_NUMBER: _ClassVar[int]
    MEMBERS_FIELD_NUMBER: _ClassVar[int]
    revision: int
    leader: str
    leaderRevision: int
    totalWeight: int
    healthy: bool
    unmapped: float
    members: _containers.RepeatedCompositeFieldContainer[Member]
    def __init__(self, revision: _Optional[int] = ..., leader: _Optional[str] = ..., leaderRevision: _Optional[int] = ..., totalWeight: _Optional[int] = ..., healthy: bool = ..., unmapped: _Optional[float] = ..., members: _Optional[_Iterable[_Union[Member, _Mapping]]] = ...) -> None: ...

class Member(_message.Message):
    __slots__ = ("hash", "nodename", "up", "enabled", "start", "end", "weight", "readPreference", "httpEndpoints", "grpcEndpoints")
    HASH_FIELD_NUMBER: _ClassVar[int]
    NODENAME_FIELD_NUMBER: _ClassVar[int]
    UP_FIELD_NUMBER: _ClassVar[int]
    IN_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    WEIGHT_FIELD_NUMBER: _ClassVar[int]
    READPREFERENCE_FIELD_NUMBER: _ClassVar[int]
    HTTPENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    GRPCENDPOINTS_FIELD_NUMBER: _ClassVar[int]
    hash: int
    nodename: str
    up: bool
    enabled: bool
    start: int
    end: int
    weight: int
    readPreference: float
    httpEndpoints: str
    grpcEndpoints: str
    def __init__(self, hash: _Optional[int] = ..., nodename: _Optional[str] = ..., up: bool = ..., enabled: bool = ..., start: _Optional[int] = ..., end: _Optional[int] = ..., weight: _Optional[int] = ..., readPreference: _Optional[float] = ..., httpEndpoints: _Optional[str] = ..., grpcEndpoints: _Optional[str] = ..., **kwargs) -> None: ...

class KeyOptValue(_message.Message):
    __slots__ = ("key", "val")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VAL_FIELD_NUMBER: _ClassVar[int]
    key: str
    val: OptValue
    def __init__(self, key: _Optional[str] = ..., val: _Optional[_Union[OptValue, _Mapping]] = ...) -> None: ...

class OptValue(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...

class KeyValue(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class StreamCSVConfig(_message.Message):
    __slots__ = ("version", "label", "uuid")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    version: int
    label: str
    uuid: bytes
    def __init__(self, version: _Optional[int] = ..., label: _Optional[str] = ..., uuid: _Optional[bytes] = ...) -> None: ...

class GenerateCSVParams(_message.Message):
    __slots__ = ("queryType", "startTime", "endTime", "windowSize", "depth", "includeVersions", "streams")
    class QueryType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ALIGNED_WINDOWS_QUERY: _ClassVar[GenerateCSVParams.QueryType]
        WINDOWS_QUERY: _ClassVar[GenerateCSVParams.QueryType]
        RAW_QUERY: _ClassVar[GenerateCSVParams.QueryType]
    ALIGNED_WINDOWS_QUERY: GenerateCSVParams.QueryType
    WINDOWS_QUERY: GenerateCSVParams.QueryType
    RAW_QUERY: GenerateCSVParams.QueryType
    QUERYTYPE_FIELD_NUMBER: _ClassVar[int]
    STARTTIME_FIELD_NUMBER: _ClassVar[int]
    ENDTIME_FIELD_NUMBER: _ClassVar[int]
    WINDOWSIZE_FIELD_NUMBER: _ClassVar[int]
    DEPTH_FIELD_NUMBER: _ClassVar[int]
    INCLUDEVERSIONS_FIELD_NUMBER: _ClassVar[int]
    STREAMS_FIELD_NUMBER: _ClassVar[int]
    queryType: GenerateCSVParams.QueryType
    startTime: int
    endTime: int
    windowSize: int
    depth: int
    includeVersions: bool
    streams: _containers.RepeatedCompositeFieldContainer[StreamCSVConfig]
    def __init__(self, queryType: _Optional[_Union[GenerateCSVParams.QueryType, str]] = ..., startTime: _Optional[int] = ..., endTime: _Optional[int] = ..., windowSize: _Optional[int] = ..., depth: _Optional[int] = ..., includeVersions: bool = ..., streams: _Optional[_Iterable[_Union[StreamCSVConfig, _Mapping]]] = ...) -> None: ...

class GenerateCSVResponse(_message.Message):
    __slots__ = ("stat", "isHeader", "row")
    STAT_FIELD_NUMBER: _ClassVar[int]
    ISHEADER_FIELD_NUMBER: _ClassVar[int]
    ROW_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    isHeader: bool
    row: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., isHeader: bool = ..., row: _Optional[_Iterable[str]] = ...) -> None: ...

class SQLQueryParams(_message.Message):
    __slots__ = ("query", "params", "role")
    QUERY_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    ROLE_FIELD_NUMBER: _ClassVar[int]
    query: str
    params: _containers.RepeatedScalarFieldContainer[str]
    role: Role
    def __init__(self, query: _Optional[str] = ..., params: _Optional[_Iterable[str]] = ..., role: _Optional[_Union[Role, _Mapping]] = ...) -> None: ...

class SQLQueryResponse(_message.Message):
    __slots__ = ("stat", "SQLQueryRow")
    STAT_FIELD_NUMBER: _ClassVar[int]
    SQLQUERYROW_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    SQLQueryRow: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., SQLQueryRow: _Optional[_Iterable[bytes]] = ...) -> None: ...

class Role(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class SetCompactionConfigParams(_message.Message):
    __slots__ = ("uuid", "CompactedVersion", "reducedResolutionRanges", "unused0")
    UUID_FIELD_NUMBER: _ClassVar[int]
    COMPACTEDVERSION_FIELD_NUMBER: _ClassVar[int]
    REDUCEDRESOLUTIONRANGES_FIELD_NUMBER: _ClassVar[int]
    UNUSED0_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    CompactedVersion: int
    reducedResolutionRanges: _containers.RepeatedCompositeFieldContainer[ReducedResolutionRange]
    unused0: int
    def __init__(self, uuid: _Optional[bytes] = ..., CompactedVersion: _Optional[int] = ..., reducedResolutionRanges: _Optional[_Iterable[_Union[ReducedResolutionRange, _Mapping]]] = ..., unused0: _Optional[int] = ...) -> None: ...

class SetCompactionConfigResponse(_message.Message):
    __slots__ = ("stat",)
    STAT_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class GetCompactionConfigParams(_message.Message):
    __slots__ = ("uuid",)
    UUID_FIELD_NUMBER: _ClassVar[int]
    uuid: bytes
    def __init__(self, uuid: _Optional[bytes] = ...) -> None: ...

class GetCompactionConfigResponse(_message.Message):
    __slots__ = ("stat", "LatestMajorVersion", "CompactedVersion", "reducedResolutionRanges", "unused0")
    STAT_FIELD_NUMBER: _ClassVar[int]
    LATESTMAJORVERSION_FIELD_NUMBER: _ClassVar[int]
    COMPACTEDVERSION_FIELD_NUMBER: _ClassVar[int]
    REDUCEDRESOLUTIONRANGES_FIELD_NUMBER: _ClassVar[int]
    UNUSED0_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    LatestMajorVersion: int
    CompactedVersion: int
    reducedResolutionRanges: _containers.RepeatedCompositeFieldContainer[ReducedResolutionRange]
    unused0: int
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., LatestMajorVersion: _Optional[int] = ..., CompactedVersion: _Optional[int] = ..., reducedResolutionRanges: _Optional[_Iterable[_Union[ReducedResolutionRange, _Mapping]]] = ..., unused0: _Optional[int] = ...) -> None: ...

class ReducedResolutionRange(_message.Message):
    __slots__ = ("Start", "End", "Resolution")
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    RESOLUTION_FIELD_NUMBER: _ClassVar[int]
    Start: int
    End: int
    Resolution: int
    def __init__(self, Start: _Optional[int] = ..., End: _Optional[int] = ..., Resolution: _Optional[int] = ...) -> None: ...

class SubscriptionUpdate(_message.Message):
    __slots__ = ("op", "uuid")
    OP_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    op: SubscriptionUpdateOp
    uuid: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, op: _Optional[_Union[SubscriptionUpdateOp, str]] = ..., uuid: _Optional[_Iterable[bytes]] = ...) -> None: ...

class SubscriptionResp(_message.Message):
    __slots__ = ("stat", "uuid", "arrowBytes")
    STAT_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    ARROWBYTES_FIELD_NUMBER: _ClassVar[int]
    stat: Status
    uuid: bytes
    arrowBytes: bytes
    def __init__(self, stat: _Optional[_Union[Status, _Mapping]] = ..., uuid: _Optional[bytes] = ..., arrowBytes: _Optional[bytes] = ...) -> None: ...



from functools import partial

import btrdb
from btrdb.conn import BTrDB


def register_serializer(conn_str=None, apikey=None, profile=None):
    """
    Register serializer for BTrDB Object
    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. `192.168.1.1:4411`.
        If set to None, will look in the environment variable `$BTRDB_ENDPOINTS`
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable `$BTRDB_API_KEY`.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        `~/.predictivegrid/credentials.yaml`.
    """
    try:
        import ray
    except ImportError:
        raise ImportError("must pip install ray to register custom serializer")
    try:
        import semver
    except ImportError:
        raise ImportError("must pip install semver to register custom serializer")

    assert (
        ray.is_initialized()
    ), "Need to call ray.init() before registering custom serializer"
    ver = semver.VersionInfo.parse(ray.__version__)
    if ver.major == 0:
        ray.register_custom_serializer(
            BTrDB,
            serializer=btrdb_serializer,
            deserializer=partial(
                btrdb_deserializer, conn_str=conn_str, apikey=apikey, profile=profile
            ),
        )
    elif ver.major == 1 and ver.minor in range(
        2, 13
    ):  # as of May 2022, latest Ray version is at 1.12
        # TODO: check different versions of ray?
        ray.util.register_serializer(
            BTrDB,
            serializer=btrdb_serializer,
            deserializer=partial(
                btrdb_deserializer, conn_str=conn_str, apikey=apikey, profile=profile
            ),
        )
    else:
        raise Exception(
            "Ray version %s does not have custom serialization. Please upgrade to >= 1.4.0"
            % ray.__version__
        )


def btrdb_serializer(_):
    """
    sererialize function
    """
    return None


def btrdb_deserializer(_, conn_str=None, apikey=None, profile=None):
    """
    deserialize function

    Parameters
    ----------
    conn_str: str, default=None
        The address and port of the cluster to connect to, e.g. `192.168.1.1:4411`.
        If set to None, will look in the environment variable `$BTRDB_ENDPOINTS`
        (recommended).
    apikey: str, default=None
        The API key used to authenticate requests (optional). If None, the key
        is looked up from the environment variable `$BTRDB_API_KEY`.
    profile: str, default=None
        The name of a profile containing the required connection information as
        found in the user's predictive grid credentials file
        `~/.predictivegrid/credentials.yaml`.
    Returns
    -------
    db : BTrDB
        An instance of the BTrDB context to directly interact with the database.
    """
    return btrdb.connect(conn_str=conn_str, apikey=apikey, profile=profile)



# tests.test_collection
# Testing for the btrdb.collection module
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: test_collection.py [] allen@pingthings.io $

"""
Testing for the btrdb.collection module
"""

##########################################################################
## Imports
##########################################################################

import pytest

from btrdb.point import RawPoint
from btrdb.utils.buffer import PointBuffer

##########################################################################
## Test Constants
##########################################################################


##########################################################################
## Initialization Tests
##########################################################################


class TestPointBuffer(object):
    def test_earliest(self):
        """
        Assert earliest returns correct key
        """
        buffer = PointBuffer(3)
        buffer[2000][0] = "horse"
        buffer[1000][0] = "zebra"
        buffer[1000][1] = "leopard"
        buffer[1000][2] = "giraffe"
        buffer[3000][0] = "pig"

        assert buffer.earliest() == 1000

    def test_earliest_if_empty(self):
        """
        Assert earliest return None
        """
        buffer = PointBuffer(3)
        assert buffer.earliest() == None

    def test_next_key_ready(self):
        """
        Assert next_key_ready returns correct key
        """
        buffer = PointBuffer(3)
        buffer.add_point(0, RawPoint(time=1000, value="zebra"))
        buffer.add_point(1, RawPoint(time=1000, value="leopard"))
        buffer.add_point(2, RawPoint(time=1000, value="giraffe"))
        buffer.add_point(0, RawPoint(time=2000, value="horse"))
        buffer.add_point(0, RawPoint(time=3000, value="pig"))
        assert buffer.next_key_ready() == 1000

        buffer = PointBuffer(2)
        buffer.add_point(0, RawPoint(time=1000, value="horse"))
        buffer.add_point(0, RawPoint(time=2000, value="zebra"))
        buffer.add_point(0, RawPoint(time=3000, value="pig"))
        buffer.add_point(1, RawPoint(time=2000, value="leopard"))
        assert buffer.next_key_ready() == 1000

    def test_next_key_ready_if_none_are(self):
        """
        Assert next_key_ready returns None if nothing is ready
        """
        buffer = PointBuffer(3)
        assert buffer.next_key_ready() == None

        buffer.add_point(0, RawPoint(time=1000, value="leopard"))
        buffer.add_point(2, RawPoint(time=1000, value="giraffe"))
        buffer.add_point(0, RawPoint(time=2000, value="horse"))
        buffer.add_point(0, RawPoint(time=3000, value="pig"))
        assert buffer.next_key_ready() == None

    def test_deactivate(self):
        """
        Assert deactivate modifies active list correctly
        """
        buffer = PointBuffer(3)
        buffer.deactivate(1)
        assert buffer.active == [True, False, True]

    def test_is_ready(self):
        """
        Assert is_ready returns correct value
        """
        buffer = PointBuffer(2)
        buffer.add_point(0, RawPoint(time=1000, value="zebra"))
        buffer.add_point(1, RawPoint(time=1000, value="giraffe"))
        buffer.add_point(0, RawPoint(time=2000, value="horse"))

        assert buffer.is_ready(1000) == True
        assert buffer.is_ready(2000) == False

        buffer.deactivate(1)
        assert buffer.is_ready(2000) == True

    def test_next_key_ready_with_inactive(self):
        """
        Assert next_key_ready returns correct key with inactive stream
        """
        buffer = PointBuffer(3)
        buffer.deactivate(1)
        buffer.add_point(0, RawPoint(time=1000, value="leopard"))
        buffer.add_point(2, RawPoint(time=1000, value="leopard"))
        buffer.add_point(0, RawPoint(time=2000, value="leopard"))
        assert buffer.next_key_ready() == 1000

        # assert first key is 500 even though it was exhausted
        buffer = PointBuffer(3)
        buffer.add_point(1, RawPoint(time=500, value="leopard"))
        buffer.deactivate(1)
        buffer.add_point(0, RawPoint(time=1000, value="leopard"))
        buffer.add_point(2, RawPoint(time=1000, value="leopard"))
        assert buffer.next_key_ready() == 500



# btrdb.utils.general
# General utilities for btrdb bindings
#
# Author:   PingThings
# Created:  Wed Jan 02 17:00:49 2019 -0500
#
# Copyright (C) 2018 PingThings LLC
# For license information, see LICENSE.txt
#
# ID: general.py [] allen@pingthings.io $

"""
General utilities for btrdb bindings
"""
from datetime import timedelta

from btrdb.utils.timez import to_nanoseconds


##########################################################################
## Functions
##########################################################################
def unpack_stream_descriptor(desc):
    """
    Returns dicts for tags and annotations found in supplied stream
    """
    tags = {}
    for tag in desc.tags:
        tags[tag.key] = tag.val.value

    anns = {}
    for ann in desc.annotations:
        anns[ann.key] = ann.val.value
    return tags, anns


##########################################################################
## Pointwidth Helpers
##########################################################################


class pointwidth(object):
    """
    A representation of a period of time described by the BTrDB tree (and used in
    aligned_windows queries). The pointwidth allows users to traverse to different
    levels of the BTrDB tree and to select StatPoints directly, vastly improving the
    performance of queries. However, because the real duration of the pointwidth is
    defined in powers of 2 (e.g. 2**pointwidth nanoseconds), the durations do not map
    to human time periods (e.g. weeks or hours).

    Parameters
    ----------
    p : int
        cast an integer to the specified pointwidth
    """

    @classmethod
    def from_timedelta(cls, delta):
        """
        Returns the closest pointwidth for the given timedelta without going over the
        specified duration. Because pointwidths are in powers of 2, be sure to check
        that the returned real duration is sufficient.
        """
        return cls.from_nanoseconds(int(delta.total_seconds() * 1e9))

    @classmethod
    def from_nanoseconds(cls, nsec):
        """
        Returns the closest pointwidth for the given number of nanoseconds without going
        over the specified duration. Because pointwidths are in powers of 2, be sure to
        check that the returned real duration is sufficient.
        """
        for pos in range(62):
            nsec = nsec >> 1
            if nsec == 0:
                break
        return cls(pos)

    def for_aligned_windows(self, start, end):
        """
        Returns the aligned_windows's first and last timestamps, as well as the number of windows,
        based on a given pointwidth set.

        Parameters
        ----------
        start : int or datetime like object
            The start time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)

        end : int or datetime like object
            The end time in nanoseconds for the range to be queried. (see
            :func:`btrdb.utils.timez.to_nanoseconds` for valid input types)


        Returns
        -------
        aligned_start: int
            First timestamp would be returned by `aligned_windows` that is inclusive of specified start
            timestamp.
        aligned_end: int
            Last timestamp would be returned by `aligned_windows` that is inclusive of specified end
            timestamp.
        n_windows: int
            The number of windows would be returned by `aligned_windows`.

        Examples
        --------
        Querying `aligned_windows` of pointwidth of 30, spaning 1 day (~24 hours).

        >>> start, end = "2016-03-01", "2016-03-02"
        >>> pointwidth(30).for_aligned_windows(start, end)
        (1456790399996657664, 1456876798632525824, 80466)
        # output timestamp's `strftime` is:
        # ['2016-02-29 23:59:59.996657664', '2016-03-01 23:59:58.632525824']

        """
        start, end = to_nanoseconds(start), to_nanoseconds(end)
        aligned_start = start - (start % self.nanoseconds)
        n_windows = (end - aligned_start) // self.nanoseconds
        aligned_end = aligned_start + (self.nanoseconds * (n_windows - 1))
        return aligned_start, aligned_end, n_windows

    def __init__(self, p):
        self._pointwidth = int(p)

    @property
    def nanoseconds(self):
        return 2**self._pointwidth

    @property
    def microseconds(self):
        return self.nanoseconds / 1e3

    @property
    def milliseconds(self):
        return self.nanoseconds / 1e6

    @property
    def seconds(self):
        return self.nanoseconds / 1e9

    @property
    def minutes(self):
        return self.nanoseconds / 6e10

    @property
    def hours(self):
        return self.nanoseconds / 3.6e12

    @property
    def days(self):
        return self.nanoseconds / 8.64e13

    @property
    def weeks(self):
        return self.nanoseconds / 6.048e14

    @property
    def months(self):
        return self.nanoseconds / 2.628e15

    @property
    def years(self):
        return self.nanoseconds / 3.154e16

    def decr(self):
        return pointwidth(self - 1)

    def incr(self):
        return pointwidth(self + 1)

    def to_timedelta(self):
        """
        Returns the timedelta of the pointwidth.
        """
        return timedelta(microseconds=self.microseconds)

    def __int__(self):
        return self._pointwidth

    def __eq__(self, other):
        return int(self) == int(other)

    def __lt__(self, other):
        return int(self) < int(other)

    def __le__(self, other):
        return int(self) <= int(other)

    def __gt__(self, other):
        return int(self) > int(other)

    def __ge__(self, other):
        return int(self) >= int(other)

    def __add__(self, other):
        return pointwidth(int(self) + int(other))

    def __sub__(self, other):
        return pointwidth(int(self) - int(other))

    def __str__(self):
        """
        Returns the pointwidth to the closest unit
        """
        if self >= 55:
            return "{:0.2f} years".format(self.years)

        if self >= 52:
            return "{:0.2f} months".format(self.months)

        if self >= 50:
            return "{:0.2f} weeks".format(self.weeks)

        if self >= 47:
            return "{:0.2f} days".format(self.days)

        if self >= 42:
            return "{:0.2f} hours".format(self.hours)

        if self >= 36:
            return "{:0.2f} minutes".format(self.minutes)

        if self >= 30:
            return "{:0.2f} seconds".format(self.seconds)

        if self >= 20:
            return "{:0.2f} milliseconds".format(self.milliseconds)

        if self >= 10:
            return "{:0.2f} microseconds".format(self.microseconds)

        return "{:0.0f} nanoseconds".format(self.nanoseconds)

    def __repr__(self):
        return "<pointwidth {}>".format(int(self))



# tests.utils.test_conversions
# Testing for btrdb convertion utilities
#
# Author:   PingThings
# Created:  Mon Dec 17 15:23:25 2018 -0500
#
# For license information, see LICENSE.txt
# ID: test_conversions.py [] allen@pingthings.io $

"""
Testing for btrdb convertion utilities
"""

##########################################################################
## Imports
##########################################################################

import json
import uuid
from datetime import datetime

import numpy as np
import pytest

from btrdb.utils.conversion import AnnotationDecoder, AnnotationEncoder, to_uuid

##########################################################################
## Test Constants
##########################################################################

EXAMPLE_UUID_STR = "07d28a44-4991-492d-b9c5-2d8cec5aa6d4"
EXAMPLE_UUID_BYTES = EXAMPLE_UUID_STR.encode("ASCII")
EXAMPLE_UUID = uuid.UUID(EXAMPLE_UUID_STR)


##########################################################################
## Conversion Tests
##########################################################################


class TestAnnotationJSON(object):
    @pytest.mark.parametrize(
        "obj, expected",
        [
            (True, "true"),
            (False, "false"),
            (None, "null"),
            (3.14, "3.14"),
            (42, "42"),
            ("foo", "foo"),
            ("a long walk on the beach", "a long walk on the beach"),
            (["a", "b", "c"], '["a", "b", "c"]'),
            ({"color": "red", "foo": 24}, '{"color": "red", "foo": 24}'),
            (datetime(2018, 9, 10, 16, 30), "2018-09-10 16:30:00.000000"),
            (
                np.datetime64(datetime(2018, 9, 10, 16, 30)),
                "2018-09-10 16:30:00.000000+0000",
            ),
            (EXAMPLE_UUID, EXAMPLE_UUID_STR),
        ],
    )
    def test_annotations_encoder(self, obj, expected):
        msg = f"did not correctly encode type {type(obj)}"
        assert json.dumps(obj, cls=AnnotationEncoder, indent=None) == expected, msg

    @pytest.mark.parametrize(
        "obj, s",
        [
            (True, "true"),
            (False, "false"),
            (None, "null"),
            (3.14, "3.14"),
            (42, "42"),
            ("foo", "foo"),
            ("foo", '"foo"'),
            ("a long walk on the beach", "a long walk on the beach"),
            ("a long walk on the beach", '"a long walk on the beach"'),
            (["a", "b", "c"], '["a", "b", "c"]'),
            ({"color": "red", "foo": 24}, '{"color": "red", "foo": 24}'),
            ("2018-09-10 16:30:00.000000", "2018-09-10 16:30:00.000000"),
            ({"uuid": EXAMPLE_UUID_STR}, f'{{"uuid": "{EXAMPLE_UUID_STR}"}}'),
        ],
    )
    def test_annotations_decoder(self, obj, s):
        msg = f"did not correctly decode type {type(obj)}"
        assert json.loads(s, cls=AnnotationDecoder) == obj, msg


class TestToUUID(object):
    def test_from_bytes(self):
        """
        Assert that `to_uuid` converts from bytes
        """
        assert to_uuid(EXAMPLE_UUID_BYTES) == EXAMPLE_UUID

    def test_from_str(self):
        """
        Assert that `to_uuid` converts from str
        """
        assert to_uuid(EXAMPLE_UUID_STR) == EXAMPLE_UUID

    def test_from_uuid(self):
        """
        Assert that `to_uuid` returns passed UUID
        """
        assert to_uuid(EXAMPLE_UUID) == EXAMPLE_UUID

    def test_raises_on_bad_data(self):
        """
        Assert that `to_uuid` raises error with bad UUID string
        """
        with pytest.raises(ValueError):
            to_uuid("bad data!!!")

    def test_raises_on_incorrect_input_type(self):
        """
        Assert that `to_uuid` raises error on wrong input class
        """
        with pytest.raises(TypeError) as exc:
            to_uuid(3.0)

        assert "Cannot convert object to UUID" in str(exc)
        assert "float" in str(exc)



import random
from math import isnan, nan
from uuid import uuid4 as new_uuid

import numpy as np
import pytest
from numpy import testing as np_test

import btrdb
import btrdb.stream
from btrdb.utils.timez import currently_as_ns

try:
    import pyarrow as pa
except ImportError:
    pa = None
try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import polars as pl
    from polars import testing as pl_test
except ImportError:
    pl = None


def test_streamset_values(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    values = ss.values()
    assert len(values) == 2
    assert len(values[0]) == len(t1)
    assert len(values[1]) == len(t2)
    assert [(p.time, p.value) for p in values[0]] == list(zip(t1, d1))
    assert [(p.time, p.value) for p in values[1]] == list(zip(t2, d2))


def test_streamset_arrow_values(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_col1 = [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0]
    expected_col2 = [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan]
    expected_schema = pa.schema(
        [
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field(str(s1.uuid), pa.float64(), nullable=False),
            pa.field(str(s2.uuid), pa.float64(), nullable=False),
        ]
    )
    values = ss.arrow_values()
    times = [t.value for t in values["time"]]
    col1 = [np.nan if isnan(v.as_py()) else v.as_py() for v in values[str(s1.uuid)]]
    col2 = [np.nan if isnan(v.as_py()) else v.as_py() for v in values[str(s2.uuid)]]
    assert times == expected_times
    assert col1 == expected_col1
    assert col2 == expected_col2
    assert expected_schema.equals(values.schema)


def test_streamset_template_schema(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    schema = pa.schema(
        [
            pa.field("t", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field("a", pa.float64(), nullable=False),
            pa.field("b", pa.float32(), nullable=False),
        ]
    )
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121, schema=schema)
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_col1 = [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0]
    expected_col2 = [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan]
    values = ss.arrow_values()
    times = [t.value for t in values["t"]]
    col1 = [np.nan if isnan(v.as_py()) else v.as_py() for v in values["a"]]
    col2 = [np.nan if isnan(v.as_py()) else v.as_py() for v in values["b"]]
    assert times == expected_times
    assert col1 == expected_col1
    assert col2 == expected_col2
    assert schema.equals(values.schema)


@pytest.mark.parametrize(
    "name_callable",
    [(None), (lambda s: str(s.uuid)), (lambda s: s.name + "/" + s.collection)],
    ids=["empty", "uu_as_str", "name_collection"],
)
def test_streamset_arrow_windows_vs_windows(conn, tmp_collection, name_callable):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=121)
        .windows(width=btrdb.utils.timez.ns_delta(nanoseconds=10))
    )
    values_arrow = ss.arrow_to_dataframe(name_callable=name_callable)
    values_arrow.index = pd.DatetimeIndex(values_arrow.index)
    values_prev = ss.to_dataframe(name_callable=name_callable).convert_dtypes(
        dtype_backend="pyarrow"
    )
    values_prev.index = pd.to_datetime(values_prev.index, utc=True)
    values_prev = values_prev.convert_dtypes(
        dtype_backend="pyarrow",
    )
    col_map = {old_col: old_col + "/mean" for old_col in values_prev.columns}
    values_prev = values_prev.rename(columns=col_map)
    assert values_arrow.equals(values_prev)


def test_streamset_arrow_windows_vs_windows_agg_all(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=121)
        .windows(width=btrdb.utils.timez.ns_delta(nanoseconds=10))
    )
    values_arrow = ss.arrow_to_dataframe(name_callable=None, agg=["all"])
    values_arrow.index = pd.DatetimeIndex(values_arrow.index)
    values_prev = ss.to_dataframe(name_callable=None, agg="all")
    values_prev = values_prev.apply(lambda x: x.astype(str(x.dtype) + "[pyarrow]"))
    values_prev = values_prev.apply(
        lambda x: x.astype("uint64[pyarrow]") if "count" in x.name else x
    )
    values_prev.index = pd.DatetimeIndex(values_prev.index, tz="UTC")
    values_prev = values_prev.convert_dtypes(dtype_backend="pyarrow")
    new_cols = ["/".join(old_col) for old_col in values_prev.columns]
    values_prev.columns = new_cols
    assert values_arrow.equals(values_prev)

    with pytest.raises(
        AttributeError,
        match="cannot provide name_callable when using 'all' as aggregate at this time",
    ):
        values_prev = ss.to_dataframe(name_callable=lambda x: str(x.uuid), agg="all")
    other_arrow_df = ss.arrow_to_dataframe(
        agg=["all", "mean", "trash"], name_callable=lambda x: str(x.uuid)
    )
    assert (
        len(
            other_arrow_df.filter(
                regex=r".*\/[min,mean,max,count,stddev]", axis=1
            ).columns
        )
        == 3 * 5
    )


@pytest.mark.parametrize(
    "name_callable",
    [(None), (lambda s: str(s.uuid)), (lambda s: s.name + "/" + s.collection)],
    ids=["empty", "uu_as_str", "name_collection"],
)
def test_streamset_arrow_aligned_windows_vs_aligned_windows(
    conn, tmp_collection, name_callable
):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=121)
        .aligned_windows(pointwidth=btrdb.utils.general.pointwidth.from_nanoseconds(10))
    )
    values_arrow = ss.arrow_to_dataframe(name_callable=name_callable)
    values_arrow.index = pd.DatetimeIndex(values_arrow.index)
    values_prev = ss.to_dataframe(
        name_callable=name_callable
    )  # .convert_dtypes(dtype_backend='pyarrow')
    values_prev = values_prev.apply(lambda x: x.astype(str(x.dtype) + "[pyarrow]"))
    values_prev = values_prev.apply(
        lambda x: x.astype("uint64[pyarrow]") if "count" in x.name else x
    )
    values_prev.index = pd.DatetimeIndex(values_prev.index, tz="UTC")
    col_map = {old_col: old_col + "/mean" for old_col in values_prev.columns}
    values_prev = values_prev.rename(columns=col_map)
    assert values_arrow.equals(values_prev)


@pytest.mark.parametrize(
    "name_callable",
    [(None), (lambda s: str(s.uuid)), (lambda s: s.name + "/" + s.collection)],
    ids=["empty", "uu_as_str", "name_collection"],
)
def test_streamset_arrow_aligned_windows_join_logic(
    conn, tmp_collection, name_callable
):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 132, 140]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=141)
        .aligned_windows(pointwidth=btrdb.utils.general.pointwidth.from_nanoseconds(8))
    )
    values_arrow = ss.arrow_to_dataframe(name_callable=name_callable)
    values_arrow.index = pd.DatetimeIndex(values_arrow.index)
    values_prev = ss.to_dataframe(
        name_callable=name_callable
    )  # .convert_dtypes(dtype_backend='pyarrow')
    values_prev = values_prev.apply(lambda x: x.astype(str(x.dtype) + "[pyarrow]"))
    values_prev = values_prev.apply(
        lambda x: x.astype("uint64[pyarrow]") if "count" in x.name else x
    )
    values_prev.index = pd.DatetimeIndex(values_prev.index, tz="UTC")
    col_map = {old_col: old_col + "/mean" for old_col in values_prev.columns}
    values_prev = values_prev.rename(columns=col_map)
    assert values_arrow.equals(values_prev)


def test_streamset_empty_arrow_values(conn, tmp_collection):
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    ss = btrdb.stream.StreamSet([s]).filter(start=100, end=200)
    values = ss.arrow_values()
    expected_schema = pa.schema(
        [
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field(str(s.uuid), pa.float64(), nullable=False),
        ]
    )
    assert [t.value for t in values["time"]] == []
    assert [v for v in values[str(s.uuid)]] == []
    assert expected_schema.equals(values.schema)


def test_streamset_to_dataframe(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    values = ss.to_dataframe()
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_col1 = [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0]
    expected_col2 = [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan]
    expected_dat = {
        tmp_collection + "/s1": expected_col1,
        tmp_collection + "/s2": expected_col2,
    }
    expected_df = pd.DataFrame(expected_dat, index=expected_times)
    assert values.equals(expected_df)


def test_arrow_streamset_to_dataframe(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    values = ss.arrow_to_dataframe()
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_times = [
        pa.scalar(v, type=pa.timestamp("ns", tz="UTC")).as_py() for v in expected_times
    ]
    expected_col1 = pa.array(
        [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0], mask=[False] * 9
    )
    expected_col2 = pa.array(
        [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan], mask=[False] * 9
    )
    expected_dat = {
        "time": expected_times,
        tmp_collection + "/s1": expected_col1,
        tmp_collection + "/s2": expected_col2,
    }
    schema = pa.schema(
        fields=[
            pa.field("time", pa.timestamp("ns", tz="UTC"), nullable=False),
            pa.field(tmp_collection + "/s1", type=pa.float64(), nullable=False),
            pa.field(tmp_collection + "/s2", type=pa.float64(), nullable=False),
        ]
    )
    expected_table = pa.Table.from_pydict(mapping=expected_dat, schema=schema)
    expected_df = expected_table.to_pandas(
        timestamp_as_object=False,
        types_mapper=pd.ArrowDtype,
    ).set_index("time")
    expected_df.index = pd.DatetimeIndex(expected_df.index, tz="UTC")
    np_test.assert_array_equal(
        values.values.astype(float), expected_df.values.astype(float)
    )


def test_arrow_streamset_to_polars(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    values = ss.arrow_to_polars()
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_times = [
        pa.scalar(v, type=pa.timestamp("ns", tz="UTC")).as_py() for v in expected_times
    ]
    expected_col1 = [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0]
    expected_col2 = [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan]
    expected_dat = {
        tmp_collection + "/s1": expected_col1,
        tmp_collection + "/s2": expected_col2,
    }
    expected_df = pd.DataFrame(
        expected_dat, index=pd.DatetimeIndex(expected_times, name="time")
    )
    expected_df_pl = pl.from_pandas(expected_df, nan_to_null=False, include_index=True)
    pl_test.assert_frame_equal(values, expected_df_pl)


@pytest.mark.parametrize(
    "name_callable",
    [(None), (lambda s: str(s.uuid)), (lambda s: s.name + "/" + s.collection)],
    ids=["empty", "uu_as_str", "name_collection"],
)
def test_streamset_arrow_polars_vs_old_to_polars(conn, tmp_collection, name_callable):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    ss = btrdb.stream.StreamSet([s1, s2]).filter(start=100, end=121)
    values_arrow = ss.arrow_to_polars()
    values_non_arrow = ss.to_polars()
    expected_times = [100, 101, 105, 106, 110, 114, 115, 119, 120]
    expected_times = [
        pa.scalar(v, type=pa.timestamp("ns", tz="UTC")).as_py() for v in expected_times
    ]
    expected_col1 = [0.0, np.nan, 1.0, np.nan, 2.0, np.nan, 3.0, np.nan, 4.0]
    expected_col2 = [np.nan, 5.0, np.nan, 6.0, 7.0, 8.0, np.nan, 9.0, np.nan]
    expected_dat = {
        tmp_collection + "/s1": expected_col1,
        tmp_collection + "/s2": expected_col2,
    }
    expected_df = pd.DataFrame(
        expected_dat, index=pd.DatetimeIndex(expected_times, tz="UTC", name="time")
    )
    expected_df_pl = pl.from_pandas(expected_df, nan_to_null=False, include_index=True)
    pl_test.assert_frame_equal(values_arrow, expected_df_pl)
    pl_test.assert_frame_equal(values_non_arrow, expected_df_pl)
    pl_test.assert_frame_equal(values_non_arrow, values_arrow)


@pytest.mark.parametrize(
    "name_callable",
    [(None), (lambda s: str(s.uuid)), (lambda s: s.name + "/" + s.collection)],
    ids=["empty", "uu_as_str", "name_collection"],
)
def test_streamset_windows_arrow_polars_vs_old_to_polars(
    conn, tmp_collection, name_callable
):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=121)
        .windows(width=btrdb.utils.timez.ns_delta(nanoseconds=10))
    )
    values_arrow_pl = ss.arrow_to_polars(name_callable=name_callable)
    values_non_arrow_pl = ss.to_polars(name_callable=name_callable)
    new_names = {
        old_col: str(old_col) + "/" + "mean"
        for old_col in values_non_arrow_pl.select(pl.col(pl.Float64)).columns
    }
    new_names["time"] = "time"
    values_non_arrow_pl = values_non_arrow_pl.rename(mapping=new_names)
    assert values_arrow_pl.equals(values_non_arrow_pl)


def test_streamset_windows_aggregates_filter(conn, tmp_collection):
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    t1 = [100, 105, 110, 115, 120]
    t2 = [101, 106, 110, 114, 119]
    d1 = [0.0, 1.0, 2.0, 3.0, 4.0]
    d2 = [5.0, 6.0, 7.0, 8.0, 9.0]
    d3 = [1.0, 9.0, 44.0, 8.0, 9.0]
    s1.insert(list(zip(t1, d1)))
    s2.insert(list(zip(t2, d2)))
    s3.insert(list(zip(t2, d3)))
    ss = (
        btrdb.stream.StreamSet([s1, s2, s3])
        .filter(start=100, end=121)
        .windows(width=btrdb.utils.timez.ns_delta(nanoseconds=10))
    )
    values_arrow_df = ss.arrow_to_dataframe(agg=["mean", "stddev"])
    values_arrow_df.index = pd.DatetimeIndex(values_arrow_df.index)
    values_non_arrow_df = ss.to_dataframe(agg="all")
    values_non_arrow_df.index = pd.DatetimeIndex(values_non_arrow_df.index, tz="UTC")
    values_non_arrow_df = values_non_arrow_df.apply(
        lambda x: x.astype(str(x.dtype) + "[pyarrow]")
    )
    values_non_arrow_df = values_non_arrow_df.apply(
        lambda x: x.astype("uint64[pyarrow]") if "count" in x.name else x
    )
    new_cols = ["/".join(old_col) for old_col in values_non_arrow_df.columns]
    values_non_arrow_df.columns = new_cols
    cols_to_drop = [
        col
        for col in values_non_arrow_df.columns
        if ("count" in col) or ("min" in col) or ("max" in col)
    ]
    values_non_arrow_df.drop(columns=cols_to_drop, inplace=True)
    assert values_arrow_df.equals(values_non_arrow_df)


def test_timesnap_backward_extends_range(conn, tmp_collection):
    sec = 10**9
    tv1 = [
        [int(0.5 * sec), 0.5],
        [2 * sec, 2.0],
    ]
    tv2 = [
        [int(0.5 * sec) - 1, 0.5],
        [2 * sec, 2.0],
    ]
    tv3 = [
        [1 * sec, 1.0],
        [2 * sec, 2.0],
    ]
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    s3 = conn.create(new_uuid(), tmp_collection, tags={"name": "s3"})
    s1.insert(tv1)
    s2.insert(tv2)
    s3.insert(tv3)
    ss = btrdb.stream.StreamSet([s1, s2, s3]).filter(
        start=1 * sec, end=3 * sec, sampling_frequency=1
    )
    values = ss.arrow_values()
    assert [1 * sec, 2 * sec] == [t.value for t in values["time"]]
    assert [0.5, 2.0] == [v.as_py() for v in values[str(s1.uuid)]]
    assert [np.nan, 2.0] == [
        np.nan if isnan(v.as_py()) else v.as_py() for v in values[str(s2.uuid)]
    ]
    assert [1.0, 2.0] == [v.as_py() for v in values[str(s3.uuid)]]


def test_timesnap_forward_restricts_range(conn, tmp_collection):
    sec = 10**9
    tv = [
        [1 * sec, 1.0],
        [2 * sec, 2.0],
        [int(2.75 * sec), 2.75],
    ]
    s = conn.create(new_uuid(), tmp_collection, tags={"name": "s"})
    s.insert(tv)
    ss = btrdb.stream.StreamSet([s]).filter(start=1 * sec, sampling_frequency=1)
    values = ss.filter(end=int(3.0 * sec)).arrow_values()
    assert [1 * sec, 2 * sec] == [t.value for t in values["time"]]
    assert [1.0, 2.0] == [v.as_py() for v in values[str(s.uuid)]]
    # Same result if skipping past end instead of to end.
    assert values == ss.filter(end=int(2.9 * sec)).arrow_values()


@pytest.mark.parametrize("freq", [(None), (30), (15), (1), (0.1)])
def test_timesnap_with_different_sampling_frequencies(freq, conn, tmp_collection):
    # 30hz data
    data_insert_freq = 30
    period_for_data = int(1 / data_insert_freq * 1e9)
    stop = btrdb.utils.timez.currently_as_ns()
    start = stop - btrdb.utils.timez.ns_delta(minutes=10)
    t1 = [i for i in range(start, stop - period_for_data, period_for_data)]
    v1 = [
        random.random() for _ in range(start, stop - period_for_data, period_for_data)
    ]
    s1 = conn.create(new_uuid(), tmp_collection, tags={"name": "s1"})
    s2 = conn.create(new_uuid(), tmp_collection, tags={"name": "s2"})
    stset = btrdb.stream.StreamSet([s1, s2])
    data_map = {s.uuid: pa.table([t1, v1], names=["time", "value"]) for s in stset}
    stset.arrow_insert(data_map=data_map)
    df = stset.filter(
        start=start, end=stop, sampling_frequency=freq
    ).arrow_to_dataframe()
    total_points = df.shape[0] * df.shape[1]
    total_raw_pts = len(v1) * len(stset)
    expected_frac_of_pts = 1 if freq is None else freq / data_insert_freq
    actual_frac_of_pts = total_points / total_raw_pts
    assert np.isclose(expected_frac_of_pts, actual_frac_of_pts, rtol=0.001)



#!/usr/bin/env python
# setup
# Setup script for installing btrdb bindings
#
# Author:   PingThings
# Created:  Mon Jan 07 14:45:32 2019 -0500
#
# For license information, see LICENSE.txt
# ID: setup.py [] allen@pingthings.io $

"""
Setup script for installing btrdb bindings.
"""

##########################################################################
## Imports
##########################################################################

import codecs
import os

from setuptools import find_packages, setup

##########################################################################
## Package Information
##########################################################################

## Basic information
NAME = "btrdb"
DESCRIPTION = "Bindings to interact with the Berkeley Tree Database using gRPC."
AUTHOR = "PingThingsIO"
EMAIL = "support@pingthings.io"
MAINTAINER = "PingThingsIO"
LICENSE = "BSD-3-Clause"
REPOSITORY = "https://github.com/PingThingsIO/btrdb-python"
PACKAGE = "btrdb"
URL = "http://btrdb.io/"
DOCS_URL = "https://btrdb-python.readthedocs.io/"

## Define the keywords
KEYWORDS = ("btrdb", "berkeley", "timeseries", "database", "bindings" "gRPC")

## Define the classifiers
## See https://pypi.python.org/pypi?%3Aaction=list_classifiers
CLASSIFIERS = (
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Database",
    "Topic :: Software Development :: Libraries :: Python Modules",
)

## Important Paths
PROJECT = os.path.abspath(os.path.dirname(__file__))
REQUIRE_PATH = "requirements.txt"
VERSION_PATH = os.path.join(PACKAGE, "version.py")
PKG_DESCRIBE = "DESCRIPTION.md"

## Directories to ignore in find_packages
EXCLUDES = (
    "tests",
    "docs",
)

##########################################################################
## Helper Functions
##########################################################################


def read(*parts):
    """
    Assume UTF-8 encoding and return the contents of the file located at the
    absolute path from the REPOSITORY joined with *parts.
    """
    with codecs.open(os.path.join(PROJECT, *parts), "rb", "utf-8") as f:
        return f.read()


def get_version(path=VERSION_PATH):
    """
    Reads the python file defined in the VERSION_PATH to find the get_version
    function, and executes it to ensure that it is loaded correctly. Separating
    the version in this way ensures no additional code is executed.
    """
    namespace = {}
    exec(read(path), namespace)
    return namespace["get_version"](short=True)


def get_requires(path=REQUIRE_PATH):
    """
    Yields a generator of requirements as defined by the REQUIRE_PATH which
    should point to a requirements.txt output by `pip freeze`.
    """
    for line in read(path).splitlines():
        line = line.strip()
        if line and not line.startswith("#"):
            yield line


def get_description_type(path=PKG_DESCRIBE):
    """
    Returns the long_description_content_type based on the extension of the
    package describe path (e.g. .txt, .rst, or .md).
    """
    _, ext = os.path.splitext(path)
    return {
        ".rst": "text/x-rst",
        ".txt": "text/plain",
        ".md": "text/markdown",
    }[ext]


##########################################################################
## Define the configuration
##########################################################################

config = {
    "name": NAME,
    "version": get_version(),
    "description": DESCRIPTION,
    "long_description": read(PKG_DESCRIBE),
    "long_description_content_type": get_description_type(PKG_DESCRIBE),
    "classifiers": list(CLASSIFIERS),
    "keywords": list(KEYWORDS),
    "license": LICENSE,
    "author": AUTHOR,
    "author_email": EMAIL,
    "maintainer": MAINTAINER,
    "maintainer_email": EMAIL,
    "project_urls": {
        "Documentation": DOCS_URL,
        "Download": "{}/tarball/v{}".format(REPOSITORY, get_version()),
        "Source": REPOSITORY,
        "Tracker": "{}/issues".format(REPOSITORY),
    },
    "download_url": "{}/tarball/v{}".format(REPOSITORY, get_version()),
    "packages": find_packages(where=PROJECT, exclude=EXCLUDES),
    "package_data": {
        "btrdb": ["grpcinterface/btrdb.proto"],
    },
    "zip_safe": False,
    "entry_points": {
        "console_scripts": [],
    },
    "install_requires": list(get_requires()),
    "python_requires": ">=3.6, <4",
    "setup_requires": ["pytest-runner"],
    "tests_require": ["pytest"],
}


##########################################################################
## Run setup script
##########################################################################

if __name__ == "__main__":
    setup(**config)



# btrdb.utils.buffer
# Module for buffering utilities
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: buffer.py [] allen@pingthings.io $

"""
Module for buffering utilities
"""

##########################################################################
## Imports
##########################################################################

from collections import defaultdict

##########################################################################
## Classes
##########################################################################


class PointBuffer(defaultdict):
    def __init__(self, length, *args, **kwargs):
        super().__init__(lambda: [None] * length, *args, **kwargs)
        self.num_streams = length
        self.active = [True] * length
        self.last_known_time = [None] * length

    def add_point(self, stream_index, point):
        self.last_known_time[stream_index] = max(
            (self.last_known_time[stream_index] or -1), point.time
        )

        self[point.time][stream_index] = point

    def earliest(self):
        if len(self.keys()) == 0:
            return None
        return min(self.keys())

    def deactivate(self, stream_index):
        self.active[stream_index] = False

    def is_ready(self, key):
        """
        Returns bool indicating whether a given key has all points from active
        (at the time) streams.
        """
        # check each stream value to see if its valid/ready
        for idx, latest_time in enumerate(self.last_known_time):
            # return False if stream is active and we are waiting on val
            if self.active[idx] and (latest_time is None or key > latest_time):
                return False

        return True

    def next_key_ready(self):
        """ """
        keys = list(self.keys())
        keys.sort()
        for key in keys:
            if self.is_ready(key):
                return key

        return None



# btrdb.conn
# Connection related objects for the BTrDB library
#
# Author:   PingThings
# Created:  Fri Dec 21 14:57:30 2018 -0500
#
# For license information, see LICENSE.txt
# ID: conn.py [] allen@pingthings.io $

"""
Connection related objects for the BTrDB library
"""

##########################################################################
## Imports
##########################################################################

import importlib.metadata
import json
import logging
import os
import re
import uuid as uuidlib
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Union
from warnings import warn

import certifi
import grpc
from grpc._cython.cygrpc import CompressionAlgorithm

from btrdb.exceptions import BTrDBError, InvalidOperation, StreamNotFoundError, retry
from btrdb.stream import Stream, StreamSet
from btrdb.utils.conversion import to_uuid
from btrdb.utils.general import unpack_stream_descriptor

##########################################################################
## Module Variables
##########################################################################

MIN_TIME = -(16 << 56)
MAX_TIME = 48 << 56
MAX_POINTWIDTH = 63


##########################################################################
## Classes
##########################################################################
logger = logging.getLogger(__name__)


class Connection(object):
    def __init__(self, addrportstr, apikey=None):
        """
        Connects to a BTrDB server

        Parameters
        ----------
        addrportstr : str, required
            The address of the cluster to connect to, e.g 123.123.123:4411
        apikey : str, optional
            The optional API key to authenticate requests

        Notes
        -----
        The ``btrdb.connect`` method is a helper function to make connecting to the platform easier
            usually that will be sufficient for most users.
        """
        # 4 Is a magic number to make sure the error propagates where btrdb.connect is called.
        warn(
            "This API is deprecated in favor of the pingthings_api, refer to your hub landing page for further documentation.",
            FutureWarning,
            stacklevel=4,
        )
        addrport = addrportstr.split(":", 2)
        # 100MB size limit ~ 2500 streams for 5000 points with each point being 64bit
        # 500MB size limit ~ 13K streams for 5000 points
        # -1 size limit = no limit of size to send
        chan_ops = [("grpc.max_receive_message_length", -1)]

        if len(addrport) != 2:
            raise ValueError("expecting address:port")

        if addrport[1] == "4411":
            # grpc bundles its own CA certs which will work for all normal SSL
            # certificates but will fail for custom CA certs. Allow the user
            # to specify a CA bundle via env var to overcome this
            env_bundle = os.getenv("BTRDB_CA_BUNDLE", "")

            # certifi certs are provided as part of this package install
            # https://github.com/certifi/python-certifi
            lib_certs = certifi.where()

            ca_bundle = env_bundle

            if ca_bundle == "":
                ca_bundle = lib_certs
            try:
                with open(ca_bundle, "rb") as f:
                    contents = f.read()
            except Exception:
                if env_bundle != "":
                    # The user has given us something but we can't use it, we need to make noise
                    raise Exception(
                        "BTRDB_CA_BUNDLE(%s) env is defined but could not read file"
                        % ca_bundle
                    )
                else:
                    contents = None

            if apikey is None:
                self.channel = grpc.secure_channel(
                    addrportstr,
                    grpc.ssl_channel_credentials(contents),
                    options=chan_ops,
                )
            else:
                self.channel = grpc.secure_channel(
                    addrportstr,
                    grpc.composite_channel_credentials(
                        grpc.ssl_channel_credentials(contents),
                        grpc.access_token_call_credentials(apikey),
                    ),
                    options=chan_ops,
                )
        else:
            self.channel = grpc.insecure_channel(addrportstr, chan_ops)
            if apikey is not None:

                class AuthCallDetails(grpc.ClientCallDetails):
                    def __init__(self, apikey, client_call_details):
                        metadata = []
                        if client_call_details.metadata is not None:
                            metadata = list(client_call_details.metadata)
                        metadata.append(("authorization", "Bearer " + apikey))
                        version = "unknown"
                        try:
                            version = importlib.metadata.version("btrdb")
                        except Exception:
                            pass
                        metadata.append(
                            (
                                "x-api-client",
                                "btrdbpy-" + version,
                            )
                        )
                        self.method = client_call_details.method
                        self.timeout = client_call_details.timeout
                        self.credentials = client_call_details.credentials
                        self.wait_for_ready = client_call_details.wait_for_ready
                        self.compression = client_call_details.compression
                        self.metadata = metadata

                class AuthorizationInterceptor(
                    grpc.UnaryUnaryClientInterceptor,
                    grpc.UnaryStreamClientInterceptor,
                    grpc.StreamUnaryClientInterceptor,
                    grpc.StreamStreamClientInterceptor,
                ):
                    def __init__(self, apikey):
                        self.apikey = apikey

                    def intercept_unary_unary(
                        self, continuation, client_call_details, request
                    ):
                        return continuation(
                            AuthCallDetails(self.apikey, client_call_details), request
                        )

                    def intercept_unary_stream(
                        self, continuation, client_call_details, request
                    ):
                        return continuation(
                            AuthCallDetails(self.apikey, client_call_details), request
                        )

                    def intercept_stream_unary(
                        self, continuation, client_call_details, request_iterator
                    ):
                        return continuation(
                            AuthCallDetails(self.apikey, client_call_details),
                            request_iterator,
                        )

                    def intercept_stream_stream(
                        self, continuation, client_call_details, request_iterator
                    ):
                        return continuation(
                            AuthCallDetails(self.apikey, client_call_details),
                            request_iterator,
                        )

                self.channel = grpc.intercept_channel(
                    self.channel,
                    AuthorizationInterceptor(apikey),
                )


def _is_arrow_enabled(info):
    info = {
        "majorVersion": info.majorVersion,
        "minorVersion": info.minorVersion,
    }
    major = info.get("majorVersion", -1)
    minor = info.get("minorVersion", -1)
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"major version: {major}")
        logger.debug(f"minor version: {minor}")
    if major >= 5 and minor >= 30:
        return True
    else:
        return False


class BTrDB(object):
    """
    The primary server connection object for communicating with a BTrDB server.
    """

    def __init__(self, endpoint):
        self.ep = endpoint
        try:
            _ = self.ep.info()
        except Exception as err:
            raise BTrDBError(f"Could not connect to the database, error message: {err}")
        self._executor = ThreadPoolExecutor()
        try:
            self._ARROW_ENABLED = _is_arrow_enabled(self.ep.info())
        except Exception:
            self._ARROW_ENABLED = False
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"ARROW ENABLED: {self._ARROW_ENABLED}")

    @retry
    def query(
        self,
        stmt: str,
        params: Union[Tuple[str], List[str]] = None,
        auto_retry=False,
        retries=5,
        retry_delay=3,
        retry_backoff=4,
    ):
        """
        Performs a SQL query on the database metadata and returns a list of
        dictionaries from the resulting cursor.

        Parameters
        ----------
        stmt : str
            a SQL statement to be executed on the BTrDB metadata.  Available
            tables are noted below.  To sanitize inputs use a `$1` style parameter such as
            `select * from streams where name = $1 or name = $2`.
        params : list or tuple
            a list of parameter values to be sanitized and interpolated into the
            SQL statement. Using parameters forces value/type checking and is
            considered a best practice at the very least.
        auto_retry : bool, default: False
            Whether to retry this request in the event of an error
        retries : int, default: 5
            Number of times to retry this request if there is an error. Will
            be ignored if auto_retry is False
        retry_delay : int, default: 3
            initial time to wait before retrying function call if there is an error.
            Will be ignored if auto_retry is False
        retry_backoff : int, default: 4
            Exponential factor by which the backoff increases between retries.
            Will be ignored if auto_retry is False

        Returns
        -------
        list
            a list of dictionary object representing the cursor results.


        Notes
        -------
        Parameters will be inserted into the SQL statement as noted by the
        parameter number such as `$1`, `$2`, or `$3`.  The `streams` table is
        available for `SELECT` statements only.

        See https://btrdb.readthedocs.io/en/latest/ for more info.

        The following are the queryable columns in the postgres ``streams`` table.

        +------------------+------------------------+-----------+
        |      Column      |          Type          | Nullable  |
        +==================+========================+===========+
        | uuid             | uuid                   | not null  |
        +------------------+------------------------+-----------+
        | collection       | character varying(256) | not null  |
        +------------------+------------------------+-----------+
        | name             | character varying(256) | not null  |
        +------------------+------------------------+-----------+
        | unit             | character varying(256) | not null  |
        +------------------+------------------------+-----------+
        | ingress          | character varying(256) | not null  |
        +------------------+------------------------+-----------+
        | property_version | bigint                 | not null  |
        +------------------+------------------------+-----------+
        | annotations      | hstore                 |           |
        +------------------+------------------------+-----------+

        Examples
        --------
        Count all streams in the platform.

        >>> conn = btrdb.connect()
        >>> conn.query("SELECT COUNT(uuid) FROM streams")
        [{'count': ...}]

        Count all streams in the collection ``foo/bar`` by passing in the variable as a parameter.

        >>> conn.query("SELECT COUNT(uuid) FROM streams WHERE collection=$1::text", params=["foo/bar"])
        [{'count': ...}]

        Count all streams in the platform that has a non-null entry for the metadata annotation ``foo``.

        >>> conn.query("SELECT COUNT(uuid) FROM streams WHERE annotations->$1::text IS NOT NULL", params=["foo"])
        [{'count': ...}]
        """
        if params is None:
            params = list()
        return [
            json.loads(row.decode("utf-8"))
            for page in self.ep.sql_query(stmt, params)
            for row in page
        ]

    def streams(self, *identifiers, versions=None, is_collection_prefix=False):
        """
        Returns a StreamSet object with BTrDB streams from the supplied
        identifiers.  If any streams cannot be found matching the identifier
        then a ``StreamNotFoundError`` will be returned.

        Parameters
        ----------
        identifiers : str or UUID
            a single item or iterable of items which can be used to query for
            streams. Identifiers are expected to be UUID as string, UUID as UUID,
            or collection/name string.

        versions : list[int]
            a single or iterable of version numbers to match the identifiers

        is_collection_prefix : bool, default=False
            If providing a collection string, is that string just a prefix, or the entire collection name?
            This will impact how many streams are returned.


        Returns
        -------
        :class:`StreamSet`
            Collection of streams.

        Examples
        --------
        With a sequence of uuids.

        >>> conn = btrdb.connect()
        >>> conn.streams(identifiers=list_of_uuids)
        <btrdb.stream.StreamSet at 0x...>

        With a sequence of uuids and version numbers.
        Here we are using version 0 to use the latest data points.

        >>> conn.streams(identifiers=list_of_uuids, versions=[0 for _ in list_of_uuids])
        <btrdb.stream.StreamSet at 0x...>

        Filtering by ``collection`` prefix ``"foo"`` where multiple collections exist like the following:
        ``foo/bar``, ``foo/baz``, ``foo/bar/new``, and ``foo``.
        If we set `is_collection_prefix`` to ``True``, this will return all streams that exist in the collections defined above.
        It is similar to a regex pattern ``^foo.*`` for matching purposes.

        >>> conn.streams(identifiers="foo", is_collection_prefix=True)
        <btrdb.stream.StreamSet at 0x...>

        If you set ``is_collection_prefix`` to ``False``, this will assume that the string identifier you provide is the full collection name.
        Matching like the regex here: ``^foo``

        >>> conn.streams(identifiers="foo", is_collection_prefix=False)
        <btrdb.stream.StreamSet at 0x...>

        """
        if versions is not None and not isinstance(versions, list):
            raise TypeError("versions argument must be of type list")

        if versions and len(versions) != len(identifiers):
            raise ValueError("number of versions does not match identifiers")
        streams: List[Stream] = []
        for ident in identifiers:
            if isinstance(ident, uuidlib.UUID):
                streams.append(self.stream_from_uuid(ident))
                continue

            if isinstance(ident, str):
                # attempt UUID lookup
                pattern = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
                if re.match(pattern, ident):
                    streams.append(self.stream_from_uuid(ident))
                    continue

                # attempt collection/name lookup
                if "/" in ident:
                    parts = ident.split("/")
                    found = self.streams_in_collection(
                        "/".join(parts[:-1]),
                        is_collection_prefix=is_collection_prefix,
                        tags={"name": parts[-1]},
                    )
                    if isinstance(found, Stream):
                        streams.append(found)
                        continue
                    if isinstance(found, list) and len(found) == 1:
                        streams.append(found[0])
                        continue
                    raise StreamNotFoundError(f"Could not identify stream `{ident}`")

            raise ValueError(
                f"Could not identify stream based on `{ident}`.  Identifier must be UUID or collection/name."
            )

        obj = StreamSet(streams)

        if versions:
            version_dict = {
                streams[idx].uuid: versions[idx] for idx in range(len(versions))
            }
            obj.pin_versions(version_dict)

        return obj

    def stream_from_uuid(self, uuid):
        """
        Creates a stream handle to the BTrDB stream with the UUID `uuid`. This
        method does not check whether a stream with the specified UUID exists.
        It is always good form to check whether the stream existed using
        `stream.exists()`.


        Parameters
        ----------
        uuid : UUID
            The uuid of the requested stream.

        Returns
        -------
        Stream
            instance of Stream class or None

        Examples
        --------

        >>> import btrdb
        >>> conn = btrdb.connect()
        >>> uuid = "f98f4b4e-9fab-46b5-8a80-f282059d69b1"
        >>> stream = conn.stream_from_uuid(uuid)
        >>> stream
        <Stream collection=foo/test name=test_stream>

        """
        return Stream(self, to_uuid(uuid))

    @retry
    def create(
        self,
        uuid,
        collection,
        tags=None,
        annotations=None,
        auto_retry=False,
        retries=5,
        retry_delay=3,
        retry_backoff=4,
    ):
        """
        Tells BTrDB to create a new stream with UUID `uuid` in `collection` with specified `tags` and `annotations`.

        Parameters
        ----------
        uuid : UUID, required
            The uuid of the requested stream.
        collection : str, required
            The collection string prefix that the stream will belong to.
        tags : dict, required
            The tags-level metadata key:value pairs.
        annotations : dict, optional
            The mutable metadata of the stream, key:value pairs
        auto_retry : bool, default: False
            Whether to retry this request in the event of an error
        retries : int, default: 5
            Number of times to retry this request if there is an error. Will
            be ignored if auto_retry is False
        retry_delay : int, default: 3
            initial time to wait before retrying function call if there is an error.
            Will be ignored if auto_retry is False
        retry_backoff : int, default: 4
            Exponential factor by which the backoff increases between retries.
            Will be ignored if auto_retry is False

        Returns
        -------
        Stream
            instance of Stream class


        Examples
        --------
        >>> import btrdb
        >>> from uuid import uuid4 # this generates a random uuid
        >>> conn = btrdb.connect()
        >>> collection = "new/stream/collection"
        >>> tags = {"name":"foo", "unit":"V"}
        >>> annotations = {"bar": "baz"}
        >>> s = conn.create(uuid=uuid4(), tags=tags, annotations=annotations, collection=collection)
        <Stream collection=new/stream/collection name=foo>
        """

        if tags is None:
            tags = {}

        if annotations is None:
            annotations = {}

        self.ep.create(uuid, collection, tags, annotations)
        return Stream(
            self,
            uuid,
            known_to_exist=True,
            collection=collection,
            tags=tags.copy(),
            annotations=annotations.copy(),
            property_version=1,
        )

    def info(self):
        """
        Returns information about the platform proxy server.

        Returns
        -------
        dict
            Proxy server connection and status information

        Examples
        --------
        >>> conn = btrdb.connect()
        >>> conn.info()
        {
        ..        'majorVersion': 5,
        ..        'minorVersion': 8,
        ..        'build': ...,
        ..        'proxy': ...,
        }

        """
        info = self.ep.info()
        return {
            "majorVersion": info.majorVersion,
            "minorVersion": info.minorVersion,
            "build": info.build,
            "proxy": {"proxyEndpoints": [ep for ep in info.proxy.proxyEndpoints]},
        }

    def list_collections(self, starts_with=""):
        """
        Returns a list of collection paths using the `starts_with` argument for
        filtering.

        Parameters
        ----------
        starts_with : str, optional, default: ''
            Filter collections that start with the string provided, if none passed, will list all collections.

        Returns
        -------
        collections: List[str]

        Examples
        --------

        Assuming we have the following collections in the platform:
        ``foo``, ``bar``, ``foo/baz``, ``bar/baz``

        >>> conn = btrdb.connect()
        >>> conn.list_collections().sort()
        ["bar", "bar/baz", "foo", "foo/bar"]

        >>> conn.list_collections(starts_with="foo")
        ["foo", "foo/bar"]


        """
        return [c for some in self.ep.listCollections(starts_with) for c in some]

    def _list_unique_tags_annotations(self, key, collection):
        """
        Returns a SQL statement and parameters to get list of tags or annotations.
        """
        if key == "annotations":
            query = "select distinct({}) as {} from streams".format(
                "skeys(annotations)", "annotations"
            )
        else:
            query = "select distinct({}) as {} from streams".format(key, key)
        params = []
        if isinstance(collection, str):
            params.append("{}%".format(collection))
            query = " where ".join([query, """collection like $1"""])
        return [metadata[key] for metadata in self.query(query, params)]

    def list_unique_annotations(self, collection=None):
        """
        Returns a list of annotation keys used in a given collection prefix.

        Parameters
        ----------
        collection : str
            Prefix of the collection to filter.

        Returns
        -------
        annotations : list[str]

        Notes
        -----
        This query treats the ``collection`` string as a prefix, so ``collection="foo"`` will match with the following wildcard syntax ``foo%``.
        If you only want to filter for a single collection, you will need to provide the full collection, if there are other collections
        that match the ``foo%`` pattern, you might need to use a custom SQL query using ``conn.query``.

        Examples
        --------
        >>> conn.list_unique_annotations(collection="sunshine/PMU1")
        ['foo', 'location', 'impedance']

        """
        return self._list_unique_tags_annotations("annotations", collection)

    def list_unique_names(self, collection=None):
        """
        Returns a list of names used in a given collection prefix.

        Parameters
        ----------
        collection : str
            Prefix of the collection to filter.

        Returns
        -------
        names : list[str]

        Examples
        --------
        Can specify a full ``collection`` name.

        >>> conn.list_unique_names(collection="sunshine/PMU1")
        ['C1ANG', 'C1MAG', 'C2ANG', 'C2MAG', 'C3ANG', 'C3MAG', 'L1ANG', 'L1MAG', 'L2ANG', 'L2MAG', 'L3ANG', 'L3MAG', 'LSTATE']

        And also provide a ``collection`` prefix.

        >>> conn.list_unique_names(collection="sunshine/")
        ['C1ANG', 'C1MAG', 'C2ANG', 'C2MAG', 'C3ANG', 'C3MAG', 'L1ANG', 'L1MAG', 'L2ANG', 'L2MAG', 'L3ANG', 'L3MAG', 'LSTATE']



        """
        return self._list_unique_tags_annotations("name", collection)

    def list_unique_units(self, collection=None):
        """
        Returns a list of units used in a given collection prefix.

        Parameters
        ----------
        collection : str
            Prefix of the collection to filter.

        Returns
        -------
        units : list[str]


        Examples
        --------

        >>> conn.list_unique_units(collection="sunshine/PMU1")
        ['amps', 'deg', 'mask', 'volts']

        """
        return self._list_unique_tags_annotations("unit", collection)

    @retry
    def streams_in_collection(
        self,
        *collection,
        is_collection_prefix=True,
        tags=None,
        annotations=None,
        auto_retry=False,
        retries=5,
        retry_delay=3,
        retry_backoff=4,
    ):
        """
        Search for streams matching given parameters

        This function allows for searching

        Parameters
        ----------
        collection : str
            collections to use when searching for streams, case sensitive.
        is_collection_prefix : bool
            Whether the collection is a prefix.
        tags : Dict[str, str]
            The tags to identify the stream.
        annotations : Dict[str, str]
            The annotations to identify the stream.
        auto_retry : bool, default: False
            Whether to retry this request in the event of an error
        retries : int, default: 5
            Number of times to retry this request if there is an error. Will
            be ignored if auto_retry is False
        retry_delay : int, default: 3
            initial time to wait before retrying function call if there is an error.
            Will be ignored if auto_retry is False
        retry_backoff : int, default: 4
            Exponential factor by which the backoff increases between retries.
            Will be ignored if auto_retry is False

        Returns
        ------
        list[Stream]
            A list of ``Stream`` objects found with the provided search arguments.


        .. note::

            In a future release, the default return value of this function will be a ``StreamSet``

        Examples
        --------

        >>> conn = btrdb.connect()
        >>> conn.streams_in_collection(collection="foo", is_collection_prefix=True)
        [<Stream collection=foo name=test1>, <Stream collection=foo name=test2,
        ... <Stream collection=foo/bar, name=testX>, <Stream collection=foo/baz/bar name=testY>]

        >>> conn.streams_in_collection(collection="foo", is_collection_prefix=False)
        [<Stream collection=foo, name=test1>, <Stream collection=foo, name=test2>]

        >>> conn.streams_in_collection(collection="foo",
        ...  is_collection_prefix=False, tags={"unit":"Volts"})
        [<Stream collection=foo, name=test1>]

        >>> conn.streams_in_collection(collection="foo",
        ...  is_collection_prefix=False, tags={"unit":"UNKNOWN"})
        []
        """
        result = []

        if tags is None:
            tags = {}

        if annotations is None:
            annotations = {}

        if not collection:
            collection = [None]

        for item in collection:
            streams = self.ep.lookupStreams(
                item, is_collection_prefix, tags, annotations
            )
            for desclist in streams:
                for desc in desclist:
                    tagsanns = unpack_stream_descriptor(desc)
                    result.append(
                        Stream(
                            self,
                            uuidlib.UUID(bytes=desc.uuid),
                            known_to_exist=True,
                            collection=desc.collection,
                            tags=tagsanns[0],
                            annotations=tagsanns[1],
                            property_version=desc.propertyVersion,
                        )
                    )
        # TODO: In future release update this method to return a streamset object.
        warn(
            "StreamSet will be the default return object for ``streams_in_collection`` in a future release.",
            FutureWarning,
            stacklevel=2,
        )
        return result

    @retry
    def collection_metadata(
        self,
        prefix,
        auto_retry=False,
        retries=5,
        retry_delay=3,
        retry_backoff=4,
    ):
        """
        Gives statistics about metadata for collections that match a
        ``prefix``.

        Parameters
        ----------
        prefix : str, required
            A prefix of the collection names to look at
        auto_retry : bool, default: False
            Whether to retry this request in the event of an error
        retries : int, default: 5
            Number of times to retry this request if there is an error. Will
            be ignored if auto_retry is False
        retry_delay : int, default: 3
            initial time to wait before retrying function call if there is an error.
            Will be ignored if auto_retry is False
        retry_backoff : int, default: 4
            Exponential factor by which the backoff increases between retries.
            Will be ignored if auto_retry is False

        Returns
        -------
        tuple
            A tuple of dictionaries containing metadata on the streams in the
            provided collection.

        Examples
        --------
        >>> conn.collection_metadata("sunshine/PMU1")
        ({'name': 0, 'unit': 0, 'ingress': 0, 'distiller': 0},
        .. {'foo': 1, 'impedance': 12, 'location': 12})

        >>> conn.collection_metadata("sunshine/")
        ({'name': 0, 'unit': 0, 'ingress': 0, 'distiller': 0},
        .. {'foo': 1, 'impedance': 72, 'location': 72})

        """
        ep = self.ep
        tags, annotations = ep.getMetadataUsage(prefix)
        pyTags = {tag.key: tag.count for tag in tags}
        pyAnn = {ann.key: ann.count for ann in annotations}
        return pyTags, pyAnn

    def __reduce__(self):
        raise InvalidOperation("BTrDB object cannot be reduced.")
